#include "ili_cache.h"
#include <iostream>
#include <fstream>
#include <nlp_macros.h>
#include <assert.h>
using namespace std;


typedef set <int>	int_set_t;

int	CacheInterface::i_ForegroundFeatureIndex = -1;
int CacheInterface::i_NewlyForegroundFeatureIndex = -1;
bool CacheInterface::b_Verbose = false;



//																
String CacheInterface::MakeEnvironmentCurrent (void)
{
	traceout << "CacheInterface::MakeEnvironmentCurrent" << endl;
	// replay previous actions to bring environment to 	
	// current expected state ...						
	CacheNode* pLastCacheNode = NULL;
	s_PreviousActionSent = "";
	s_PreviousActionRequested = "";
	s_LastLiveState = "";

	ITERATE (HistoryInCacheNodes_dq_t, dq_HistoryFromCache, ite)
	{
		String sAction = ite->first;
		CacheNode* pNextCacheNode = ite->second;

		String sConvertedAction = ConvertActionToLatest (sAction, pLastCacheNode);
		if ("" == sConvertedAction)
			return "";

		s_LastLiveState = SendReceiveDirect (sConvertedAction, sAction);
		if ("" == s_LastLiveState)
			return "";

		pLastCacheNode = pNextCacheNode;
		s_PreviousActionSent = sConvertedAction;
		s_PreviousActionRequested = sAction;
	}

	return s_LastLiveState;
}




//																
String CacheInterface::ConvertActionToLatest (String& _sAction, 
											  CacheNode* _pCacheNode)
{
	String x (_sAction);
	x.Strip ();
	traceout << "CacheInterface::ConvertActionToLatest (" << x << ")" << endl;

	if ('!' != _sAction.at (0))
		return _sAction;

	assert (NULL != _pCacheNode);
	String sConvertedAction;

	int iAttempts = 0;
	while (iAttempts < 2)
	{
		++ iAttempts;

		bool bLogError = (2 == iAttempts);
		sConvertedAction = UpdateActionFromCache (_sAction, 
												  s_LastLiveState,
												  _pCacheNode,
												  bLogError);
		if ("" != sConvertedAction)
			return sConvertedAction;

		if (("" != s_PreviousActionSent) && ('*' == s_PreviousActionSent.at (0)))
		{
			String sStrippedAction (s_PreviousActionSent);
			sStrippedAction.Strip ();
			cerr << "[INFO] Will retry request. Attempt : " << iAttempts
				 << ", action : " << sStrippedAction
				 << endl;
			sleep (1);

			s_LastLiveState = SendReceiveDirect (s_PreviousActionSent, 
												 s_PreviousActionRequested, 
												 true);
			if ("" == s_LastLiveState)
			{
				sConvertedAction = "";
				break;
			}
			continue;
		}
		break;
	}

	// The last attempt to match up the requested action with 	
	// cache failed.  This means we have a saved state in the	
	// cache that has repeatedly conflicted against the 		
	// corresponding live state.  So we need to trim that cache	
	// branch, and it needs to be trimmed one action earlier,	
	// since that would have been the action that resulted in	
	// We mark the need to do this here using the		
	// hs_cache_mismatch flag, and then do the trimming in		
	// SaveHistoryToCache ().									
	if ("" == sConvertedAction)
	{
		cerr << "[ERROR]  Failed to convert action to latest after " << iAttempts << " attempts." << endl;
		traceout << "convert action to latest failed." << endl;

		// assert (dq_StateOfHistoryToSave.size () > 2);
		if (dq_StateOfHistoryToSave.size () >= 2)
			dq_StateOfHistoryToSave [dq_StateOfHistoryToSave.size () - 2] = hs_cache_mismatch;
		else if (dq_StateOfHistoryToSave.size () == 1)
			dq_StateOfHistoryToSave [0] = hs_cache_mismatch;
		else
		{
			cout << "[ERROR]  Encountered a cache mismatch on the first state measurement.\n"
				    "         This should not be possible!\n";
			abort ();
		}
	}

	return sConvertedAction;
}




//																
String CacheInterface::UpdateActionFromCache (String& _sAction, 
												String& _sLastLiveState,
												CacheNode* _pCacheNode,
												bool _bLogError)
{
	String x (_sAction);
	x.Strip ();
	traceout << "CacheInterface::UpdateActionFromCache (" << x << endl;
	if ('!' != _sAction.at (0))
		return _sAction;

	if (_sLastLiveState == _pCacheNode->GetState ())
	 	return _sAction;


	// extract object id from message				
	size_t iIdEnd = _sAction.find (':');
	assert (string::npos != iIdEnd);
	String sObjectIdFromMessage (_sAction.substr (1, iIdEnd - 1));	
	if ("*" == sObjectIdFromMessage)
		return _sAction;


	// find object to interact with...				
	EnvObject* pCachedObject = NULL;
	unsigned long lCachedObjectCount = 0;

	{
		// ITERATE (EnvObject_dq_t, _pCacheNode->dq_Objects, ite)
		for (int i = 0; i < _pCacheNode->GetObjectCount (); ++ i)
		{
			// EnvObject* pObject = *ite;
			EnvObject* pObject = _pCacheNode->GetObject (i);

			++ lCachedObjectCount;
			if (pObject->z_ID == sObjectIdFromMessage)
			{
				if (NULL != pCachedObject)
				{
					// size_t iNameStart = pCachedObject->s_Features.Find ("\x01");
					// size_t iFeaturesStart = pCachedObject->s_Features.Find ("\x01", iNameStart + 1);
					// String sCachedObjectFeatures =  pCachedObject->s_Features.substr (iFeaturesStart);
					String sCachedObjectFeatures;
					sCachedObjectFeatures << pCachedObject->s_TemporallyUniqueId << '\x01' 
										  << pCachedObject->bit_Features;

					// iNameStart = pObject->s_Features.Find ("\x01");
					// iFeaturesStart = pObject->s_Features.Find ("\x01", iNameStart + 1);
					// String sObjectFeatures =  pObject->s_Features.substr (iFeaturesStart);
					String sObjectFeatures;
					sObjectFeatures << pObject->s_TemporallyUniqueId << '\x01' << pObject->bit_Features;

					// if (sCachedObjectFeatures != sObjectFeatures)
					if (pCachedObject->bit_Features != pObject->bit_Features)
					{
						cerr << "[WARNING] Multiple objects with same id "
								"(this is possible due to intentional object duplication) id : "
							 << sObjectIdFromMessage << '\n'
							 << "----------------------------------" << '\n'
							 // << pCachedObject->s_ObjectData << '\n'
							 // << pCachedObject->s_Features << '\n'
							 // << "----------------------------------" << '\n'
							 // << pObject->s_ObjectData << '\n'
							 // << pObject->s_Features << '\n'
							 // << "==================================" << '\n'
							 << sCachedObjectFeatures << '\n'
							 << "----------------------------------" << '\n'
							 << sObjectFeatures << '\n'
							 << "----------------------------------" << endl;

						ofstream file ("error.log", ios::out | ios::app);
						file << "[WARNING] Multiple objects with same id "
								"(this is possible due to intentional object duplication) id : "
							 << sObjectIdFromMessage << '\n'
							 << "----------------------------------" << '\n'
							 // << pCachedObject->s_ObjectData << '\n'
							 // << pCachedObject->s_Features << '\n'
							 // << "----------------------------------" << '\n'
							 // << pObject->s_ObjectData << '\n'
							 // << pObject->s_Features << '\n'
							 // << "==================================" << '\n'
							 << sCachedObjectFeatures << '\n'
							 << "----------------------------------" << '\n'
							 << sObjectFeatures << '\n'
							 << "----------------------------------" << "\n\n\n";
						file.close ();
					}
				}
				pCachedObject = pObject;
			}
		}
	}


	if (NULL == pCachedObject)
	{
		if ((true == _bLogError) || (true == b_Verbose))
		{
			cerr << "[ERROR] Object originally selected from cache not present in cache on the next step!!"
				 << endl;
			ofstream file ("error.log", ios::out | ios::app);
			file << "[ERROR] Object originally selected from cache not present in cache on the next step!!"
				 << "\n\n\n";
			file.close ();
		}
		return "";
	}


	// convert live message to object list, and 	
	// find current id of object to interact with.	
	String sLiveObjectID;
	unsigned long lLiveObjectCount = 0;
	EnvObject* pObjectWithMessageID = NULL;

	void* pLiveObjectArray = NULL;
	int iLiveObjectCount = 0;
	// EnvObject_dq_t dqLiveObjects = p_Cache->UnpackState (_sLastLiveState);
	p_Cache->UnpackState (_sLastLiveState, &pLiveObjectArray, &iLiveObjectCount);
	{
		// ITERATE (EnvObject_dq_t, dqLiveObjects, ite)
		for (int i = 0; i < iLiveObjectCount; ++ i)
		{
			// EnvObject* pLiveObject = *ite;
			EnvObject* pLiveObject = CacheNode::GetObject (pLiveObjectArray, i);

			++ lLiveObjectCount;

			// if (pLiveObject->s_Features == pCachedObject->s_Features)
			if ((pLiveObject->s_TemporallyUniqueId == pCachedObject->s_TemporallyUniqueId) &&
				 (pLiveObject->bit_Features == pCachedObject->bit_Features))
			{
				if ("" == sLiveObjectID)
					sLiveObjectID = pLiveObject->z_ID;
				else if (pLiveObject->z_ID == sObjectIdFromMessage)
					sLiveObjectID = pLiveObject->z_ID;
			}

			if (pLiveObject->z_ID == sObjectIdFromMessage)
				pObjectWithMessageID = pLiveObject;
		}
	}

	// if we were unable to find the object to 		
	// interact with through features only, check	
	// for object id match ...						
	if (("" == sLiveObjectID) && (NULL != pObjectWithMessageID))
	{
		int_set_t	setMismatchIndex;

		// -- OBSOLETE START -------------------------------------------------------
		// The following case of a mismatch between feature counts of objects		
		// can no longer occur since the object itself checks for feature count		
		// errors on creation...													
		// size_t iFeatureCount = pCachedObject->vec_Features.Size ();
		// if (iFeatureCount != pObjectWithMessageID->vec_Features.Size ())
		// {
		// 	if ((true == _bLogError) || (true == b_Verbose))
		// 	{
		// 		cerr << "[ERROR] feature vector size mismatch between cached "
		// 				"and live? objects!" << endl
		// 			 << "        cached : " << iFeatureCount 
		// 			 << ", live : " << pObjectWithMessageID->vec_Features.Size () << endl;
		// 		cerr << "        id     : " << sObjectIdFromMessage << endl;
		// 		cerr << "        cached : " << pCachedObject->s_Features << endl;
		// 		cerr << "        live?  : " << pObjectWithMessageID->s_Features << endl;
		//
		// 		ofstream file ("error.log", ios::out | ios::app);
		// 		file << "[ERROR] feature vector size mismatch between cached "
		// 				"and live? objects!" << '\n'
		// 			 << "        cached : " << iFeatureCount 
		// 			 << ", live : " << pObjectWithMessageID->vec_Features.Size () << '\n';
		// 		file << "        id     : " << sObjectIdFromMessage << '\n';
		// 		file << "        cached : " << pCachedObject->s_Features << '\n';
		// 		file << "        live?  : " << pObjectWithMessageID->s_Features << '\n';
		// 		file << "\n\n";
		// 		file.close ();
		// 	}
		// }
		// assert (iFeatureCount == pObjectWithMessageID->vec_Features.Size ());
		// -- OBSOLETE END ---------------------------------------------------------

		for (size_t i = 0; i < FEATURE_COUNT; ++ i)
		{
			// if (pCachedObject->vec_Features [i] != pObjectWithMessageID->vec_Features [i])
			if (pCachedObject->bit_Features [i] != pObjectWithMessageID->bit_Features [i])
				setMismatchIndex.insert (i);
		}
		setMismatchIndex.erase (i_ForegroundFeatureIndex);
		setMismatchIndex.erase (i_NewlyForegroundFeatureIndex);

		if (true == setMismatchIndex.empty ())
			sLiveObjectID = sObjectIdFromMessage;
		else
		{
			if ((true == _bLogError) || (true == b_Verbose))
			{
				cerr << "[ERROR] feature mismatch, object ids match, but feature mismatch is not innocuous."
					 << endl;

				ofstream file ("error.log", ios::out | ios::app);
				file << "[ERROR] feature mismatch, object ids match, but feature mismatch is not innocuous."
					 << "\n\n\n";
				file.close ();
			}
		}
	}


	// cleanup memory...							
	if (NULL != pLiveObjectArray)
		((EnvObject*)pLiveObjectArray)->DestroyObjects (pLiveObjectArray);
	iLiveObjectCount = 0;
	// ITERATE (EnvObject_dq_t, dqLiveObjects, ite)
	// 	delete *ite;


	// check for errors ...							
	if ("" == sLiveObjectID)
	{
		if ((true == _bLogError) || (true == b_Verbose))
		{
			cerr << "[ERROR] Could not find matching live object.  Action : " << _sAction << endl;
			ofstream file ("error.log", ios::out | ios::app);
			file << "[ERROR] Could not find matching live object.\n";
			file << "Environment   : " << p_EnvironmentInterface->GetEnvironmentId () << "\n";
			file << "Action        : " << _sAction << "\n";
			file << "Cached object : " << pCachedObject->z_ID << "  " << pCachedObject->bit_Features << '\n';
			file << "-- cache -------------------------------------------" << '\n';
			file << _pCacheNode->GetState () << '\n';
			file << "----------------------------------------------------" << '\n';
			file << "-- live --------------------------------------------" << '\n';
			file << _sLastLiveState << '\n';
			file << "----------------------------------------------------" << "\n\n\n";
			file.close ();
		}
		return "";
	}
	if (lLiveObjectCount != lCachedObjectCount)
	{
		if ((true == _bLogError) || (true == b_Verbose))
		{
			cerr << "[WARNING] object count mismatch. L: " << lLiveObjectCount 
				 << ", C: " << lCachedObjectCount << endl;

			ofstream file ("error.log", ios::out | ios::app);
			file << "[WARNING] object count mismatch. L: " << lLiveObjectCount 
				 << ", C: " << lCachedObjectCount << '\n';
			file << "-- cache -------------------------------------------" << '\n';
			file << _pCacheNode->GetState () << '\n';
			file << "----------------------------------------------------" << '\n';
			file << "-- live --------------------------------------------" << '\n';
			file << _sLastLiveState << '\n';
			file << "----------------------------------------------------" << '\n';
			file << "\n\n";
			file.close ();
		}
		return "";
	}

	// convert object id to new environment ...		
	String sConvertedAction;
	sConvertedAction << '!' << sLiveObjectID << _sAction.substr (_sAction.find (':'));
	sConvertedAction.Strip ();
	sConvertedAction << '\n';
	
	return sConvertedAction;
}



//																
bool CacheInterface::SwitchFromCacheToDirect (void)
{
	traceout << "CacheInterface::SwitchFromCacheToDirect" << endl;
	b_SendReceiveDirect = true;
	p_StartingNode = NULL;
	p_CurrentNode = NULL;

	// remove reader lock since we won't be 		
	// interacting with the cache until the end of	
	// this episode.								
	assert (CACHE_LOCK_READ == c_CacheLockMode);
	p_Cache->Unlock ();
	c_CacheLockMode = CACHE_LOCK_NONE;
	
	// make environment current						
	if (false == p_EnvironmentInterface->StartEpisode (s_RequiredEnvStatingState))
		return false;
	if (false == dq_HistoryFromCache.empty ())
	{
		s_LastLiveState = MakeEnvironmentCurrent ();
		if ("" == s_LastLiveState)
			return false;
	}
	return true;
}



//																
bool CacheInterface::SaveHistoryToCache (void)
{
	traceout << "CacheInterface::SaveHistoryToCache" << endl;
	// replay cache actions to bring it to the		
	// point where the new state/actions need		
	// to be inserted ...							
	CacheNode* pRoot = p_Cache->GetRoot ();
	CacheNode* pNode = pRoot->FindChild (s_RequiredEnvStatingState);
	if (NULL == pNode)
	{
		String sEmptyState (s_RequiredEnvStatingState);
		pNode = p_Cache->CreateNode (sEmptyState);
		pRoot->AddChild (s_RequiredEnvStatingState, pNode);
	}

	History_dq_t::iterator	iteToSave = dq_HistoryToSave.begin ();
	History_dq_t::iterator	iteToSave_end = dq_HistoryToSave.end ();

	HistoryState_dq_t::iterator	iteToSaveState = dq_StateOfHistoryToSave.begin ();
	HistoryState_dq_t::iterator iteToSaveState_end = dq_StateOfHistoryToSave.end ();

	HistoryInCacheNodes_dq_t::iterator	iteFromCache = dq_HistoryFromCache.begin ();
	HistoryInCacheNodes_dq_t::iterator	iteFromCache_end = dq_HistoryFromCache.end ();

	// skip over stata/action pairs that are already in cache 
	for (; (iteFromCache != iteFromCache_end) && 
		   (iteToSave != iteToSave_end) && 
		   (iteToSaveState != iteToSaveState_end);
		 ++ iteFromCache, ++ iteToSave, ++ iteToSaveState)
	{
		String& rAction = iteFromCache->first;

		HistoryState_e eState = *iteToSaveState;
		if (hs_comm_error == eState)
			return false;

		if (hs_cache_mismatch == eState)
		{
			cout << "[INFO]  Discarding branch on cache-live mismatch." << endl;
			pNode->DiscardBranch (rAction);
			return false;
		}

		CacheNode* pNextNode = pNode->FindChild (rAction);
		// This is a race condition check.  These are actions	
		// that were originally executed against the cache, so	
		// we can expect them to be present when we get here.	
		// But if a different lerner executed a similar			
		// sequence, a live-cache mismatch was encountered, and	
		// the branch got discarded, we might get a NULL child.	
		// If this happens, it's best we just drop the trace,	
		// and let a future interaction save it...				
		if (NULL == pNextNode)
			return false;
		pNode = pNextNode;
	}

	// insert new state/action pairs							
	int iStepsIntoSave = 0;
	for (; iteToSave != iteToSave_end; ++ iteToSave, ++ iteToSaveState)
	{
		++ iStepsIntoSave;
		String& rActionSent = iteToSave->first.first;
		String& rActionRequested = iteToSave->first.second;
		String& rState = iteToSave->second;
		HistoryState_e eState = *iteToSaveState;


		// If there was a communications error, we shouldn't	
		// try to save beyond that point...						
		if (hs_comm_error == eState)
			return false;

		// If there was a persistent mismatch of a cache node	
		// with it's corresponding live state, trim the branch,	
		// and stop - there will be no valid trace beyond that.	
		if (hs_cache_mismatch == eState)
		{
			cout << "[INFO]  Discarding branch on cache-live mismatch." << endl;
			pNode->DiscardBranch (rActionRequested);
			pNode->DiscardBranch (rActionSent);
			return false;
		}

		CacheNode* pNextNodeForRequested = SaveActionStateToNode (pNode, 
																  rActionRequested,
																  rState,
																  iStepsIntoSave);
		if (NULL == pNextNodeForRequested)
			return false;

		CacheNode* pNextNodeForSent = SaveActionStateToNode (pNode,
															 rActionSent,
															 rState,
															 iStepsIntoSave);
		if (NULL == pNextNodeForSent)
			return false;

		if (pNextNodeForSent != pNextNodeForRequested)
		{
			bool bRequestedObserve = rActionRequested.StartsWith ("*");
			bool bSentObserve = rActionSent.StartsWith ("*");
			if ((true == bRequestedObserve) && (true == bSentObserve))
			{
				cerr << "[ERROR] Nodes for sent & requested mismatch!" << endl;
				cerr << "        requested : " << rActionRequested;
				cerr << "        sent      : " << rActionSent;
				if (pNextNodeForSent->GetNormalizedState () !=
						pNextNodeForRequested->GetNormalizedState ())
					cerr << "[ERROR] Node states for sent & requested mismatch!" << endl;
				else
					cerr << "[ERROR] Node states for sent & requested match!" << endl;

				ofstream file ("error.log", ios::out | ios::app);
				file << "[ERROR] Nodes for sent & requested mismatch!\n";
				file << "        requested : " << rActionRequested << '\n';
				file << "        sent      : " << rActionSent << '\n';
				file << "-- requested --------------------------------\n";
				file << pNextNodeForRequested->GetNormalizedState () << '\n';
				file << "---------------------------------------------\n";
				file << "-- sent -------------------------------------\n";
				file << pNextNodeForSent->GetNormalizedState () << '\n';
				file << "---------------------------------------------\n";
				file.close ();
			}
			else if ((true == bRequestedObserve) || (true == bSentObserve))
			{
				cerr << "[ERROR] actions for sent & requested mismatch!" << endl;
				cerr << "        requested : " << rActionRequested;
				cerr << "        sent      : " << rActionSent;
				cerr << __FILE__ << ':' << __LINE__ << "  " << __PRETTY_FUNCTION__ << endl;
				abort ();
			}
		}

		pNode = pNextNodeForRequested;
	}

	return true;
}




//																
CacheNode* CacheInterface::SaveActionStateToNode (CacheNode* _pNode, 
												  String& _rAction,
												  String& _rState,
												  int _iStepsIntoSave)
{
	CacheNode* pOldNode = _pNode->FindChild (_rAction);
	CacheNode* pNextNode = _pNode->GetNode (_rAction, _rState);

	// We get a NULL return only if we've previously taken	
	// the same action (SendAction, not MeasureState) , and 
	// gotten a different different	env response...			
	if (NULL == pNextNode)
	{
		_pNode->DiscardBranch (_rAction);
		return NULL;
	}

	if (pNextNode != pOldNode)
	{
		// If we get here, we have two measure states 		
		// returning two different env states...			
		if (NULL != pOldNode)
		{
			ofstream file ("error.log", ios::out | ios::app);
			file << "[WARNING]  node mismatch : " << (void*)pNextNode << " <> " 
				 << (void*)pOldNode << '\n';
			file << "history from cache : " << dq_HistoryFromCache.size () << '\n'
				 << "history to save    : " << dq_HistoryToSave.size () << '\n'
				 << "state of history   : " << dq_StateOfHistoryToSave.size () << '\n'
				 << "steps into save    : " << _iStepsIntoSave << '\n';
			file << "         action : " << _rAction << '\n';
			file << "-- starting state -----------------------------" << '\n';
			file << _pNode->GetState () << '\n';
			file << "-----------------------------------------------" << '\n';
			file << "-- old ----------------------------------------" << '\n';
			file << pOldNode->GetState () << '\n';
			file << "-----------------------------------------------" << '\n';
			file << "-- old normalized -----------------------------" << '\n';
			file << pOldNode->GetNormalizedState () << '\n';
			file << "-----------------------------------------------" << '\n';
			file << "-- new ----------------------------------------" << '\n';
			file << pNextNode->GetState () << '\n';
			file << "-----------------------------------------------" << '\n';
			file << "-- new normalized -----------------------------" << '\n';
			file << pNextNode->GetNormalizedState () << '\n';
			file << "-----------------------------------------------" << '\n';

			if (pOldNode->GetNormalizedState () == pNextNode->GetNormalizedState ())
				file << "nodes have identical state." << '\n';
			else
				file << "nodes don't have identical state." << '\n';


			cerr << "[WARNING]  node mismatch : " << (void*)pNextNode << " <> " 
				 << (void*)pOldNode << ", trimming cache branch." << endl;

			_pNode->DiscardBranch (_rAction);
			return NULL;
		}

		_pNode->AddChild (_rAction, pNextNode);
	}

	return pNextNode;
}



//																
CacheNode* CacheInterface::SendReceiveThroughCache (String& _sAction)
{
	assert (CACHE_LOCK_READ == c_CacheLockMode);
	CacheNode* pNextNode = p_CurrentNode->FindChild (_sAction);
	if (NULL == pNextNode)
		return NULL;

	p_LastCacheNode = pNextNode;
	p_CurrentNode = pNextNode;
	dq_HistoryFromCache.push_back (make_pair (_sAction, pNextNode));

	return pNextNode;
}



//																
String CacheInterface::SendReceiveDirect (String& _sActionToSend, 
										  String& _sActionReqested,
										  bool _bRetryAttempt)
{
	String x (_sActionToSend);
	x.Strip ();
	traceout << "CacheInterface::SendReceiveDirect (" << x << ')' << endl;

	assert (CACHE_LOCK_NONE == c_CacheLockMode);
	String sState;
	bool bRet = p_EnvironmentInterface->SendReceive (_sActionToSend, &sState);

	// This is to catch an odd edge case in the Windows2000	
	// environment when some actions result in the windows	
	// network subsystem to drop the active connection to 	
	// the learner's OS agent...							
	if ("restart complete.\x10\n" == sState)
	{
		if (true == _sActionToSend.StartsWith ("*"))
		{
			cout << "[INFO]  '* -> restart complete.  Resending." << endl;
			sState = "";
			bRet = p_EnvironmentInterface->SendReceive (_sActionToSend, &sState);
		}
		else if (true == _sActionToSend.StartsWith ("!"))
		{
			cout << "[INFO]  '! -> restart complete.  Replacing with 'action complete. env changed'." << endl;
			sState = "action completed, env has changed.\x10\n";
		}
	}

	//														
	if (true == _bRetryAttempt)
	{
		assert (false == dq_HistoryToSave.empty ());
		assert (false == dq_StateOfHistoryToSave.empty ());
		assert (dq_HistoryToSave.size () == dq_StateOfHistoryToSave.size ());

		dq_HistoryToSave [dq_HistoryToSave.size () - 1]
				= make_pair (make_pair (_sActionToSend, _sActionReqested), sState);
		dq_StateOfHistoryToSave [dq_StateOfHistoryToSave.size () - 1]
				= (true == bRet)? hs_retry : hs_comm_error;
	}
	else
	{
		dq_HistoryToSave.push_back (make_pair (make_pair (_sActionToSend, _sActionReqested), sState));
		dq_StateOfHistoryToSave.push_back ((true == bRet)? hs_ok : hs_comm_error);
	}

	return (true == bRet)? sState : "";
}





//																
CacheInterface::CacheInterface (void)
{
	c_CacheLockMode = CACHE_LOCK_NONE;

	p_Cache = NULL;
	p_EnvironmentInterface = NULL;

	s_RequiredEnvStatingState = "";

	p_StartingNode = NULL;
	p_CurrentNode = NULL;

	s_PreviousActionSent = "";
	s_PreviousActionRequested = "";
	s_LastLiveState = "";

	b_InEpisode = false;
	b_SendReceiveDirect = false;
	b_KnownWorldOnly = false;
}




//																
CacheInterface::~CacheInterface (void)
{
	c_CacheLockMode = CACHE_LOCK_NONE;
}




//																
void CacheInterface::Init (Cache* _pCache, EnvironmentInterface* _pEnvInterface)
{
	p_Cache = _pCache;
	p_EnvironmentInterface = _pEnvInterface;
	b_SendReceiveDirect = false;
}



//																
bool CacheInterface::StartEpisode (String& _sStartingState)
{
	String x (_sStartingState);
	x.Strip ();
	traceout << "CacheInterface::StartEpisode (" << x << ')' << endl;

	assert ("" == s_RequiredEnvStatingState);
	s_RequiredEnvStatingState = _sStartingState;
	b_SendReceiveDirect = false;

	assert (true == dq_HistoryToSave.empty ());
	assert (true == dq_StateOfHistoryToSave.empty ());
	assert (true == dq_HistoryFromCache.empty ());
	dq_HistoryToSave.clear ();
	dq_StateOfHistoryToSave.clear ();
	dq_HistoryFromCache.clear ();

	// lock cache for reading.				
	assert (CACHE_LOCK_NONE == c_CacheLockMode);
	p_Cache->LockForReading ();
	c_CacheLockMode = CACHE_LOCK_READ;
	
	CacheNode* pRoot = p_Cache->GetRoot ();
	p_CurrentNode = pRoot->FindChild (s_RequiredEnvStatingState);
	if (NULL == p_CurrentNode)
		return SwitchFromCacheToDirect ();

	return true;
}


//																
bool CacheInterface::EndEpisode (void)
{
	traceout << "CacheInterface::EndEpisode" << endl;
	if ("" == s_RequiredEnvStatingState)
		return false;

	// If we switched to direct interaction	with	
	// the environment, the reader lock has already	
	// been removed; otherwise, unlock it.			
	// We also need to save new interaction info	
	// to the cache.								
	bool bTraceSaved = true;
	if (true == b_SendReceiveDirect)
	{
		p_EnvironmentInterface->EndEpisode ();

		// lock cache for writing.			
		assert (CACHE_LOCK_NONE == c_CacheLockMode);
		p_Cache->LockForWriting ();
		c_CacheLockMode = CACHE_LOCK_WRITE;

		// update cache with history.		
		bTraceSaved = SaveHistoryToCache ();

		// release write lock on cache.		
		assert (CACHE_LOCK_WRITE == c_CacheLockMode);
		p_Cache->Unlock ();
		c_CacheLockMode = CACHE_LOCK_NONE;
	}
	else
	{
		// release reader lock on cache.	
		assert (CACHE_LOCK_READ == c_CacheLockMode);
		p_Cache->Unlock ();
		c_CacheLockMode = CACHE_LOCK_NONE;
	}

	dq_HistoryToSave.clear ();
	dq_StateOfHistoryToSave.clear ();
	dq_HistoryFromCache.clear ();

	s_RequiredEnvStatingState = "";
	b_SendReceiveDirect = false;

	return bTraceSaved;
}



//																
String CacheInterface::SendReceive (String& _sAction)
{
	if ("" == s_RequiredEnvStatingState)
	{
		cerr << "[ERROR]  Action request received before StartEpisode request : "
			 << _sAction << endl;
		return "[[ERROR]]  Action request received before StartEpisode request.";
	}

	String sAction (_sAction);
	String sState;
	if (false == b_SendReceiveDirect)
	{
		CacheNode* pState = SendReceiveThroughCache (sAction);
		if (NULL != pState)
		{
			sState = pState->GetState ();
			return sState;
		}

		// if b_KnownWorldOnly is set, we never 
		// interact with the actual environment	
		if (true == b_KnownWorldOnly)
			return "\x10\n";

		if (false == SwitchFromCacheToDirect ())
			return "[[ERROR]] Cache error\x10\n";

		sAction = ConvertActionToLatest (sAction, p_LastCacheNode);
		if ("" == sAction)
			return "[[ERROR]] Cache error\x10\n";
	}

	sState = SendReceiveDirect (sAction, _sAction);
	return ("" != sState)?  sState : "[[ERROR]] Cache error\x10\n";
}



//																
CacheNode* CacheInterface::SendReceive (String& _sAction, CacheNode_dq_t* _pdqNewNodes)
{
	if ("" == s_RequiredEnvStatingState)
	{
		cerr << "[ERROR]  Action request received before StartEpisode request : "
			 << _sAction << endl;
		return NULL;
	}

	// check cache first...									
	String sAction (_sAction);
	if (false == b_SendReceiveDirect)
	{
		CacheNode* pState = SendReceiveThroughCache (sAction);
		if (NULL != pState)
			return pState;

		// if b_KnownWorldOnly is set, we never 
		// interact with the actual environment	
		if (true == b_KnownWorldOnly)
			return NULL;

		if (false == SwitchFromCacheToDirect ())
			return NULL;

		sAction = ConvertActionToLatest (sAction, p_LastCacheNode);
		if ("" == sAction)
			return NULL;
	}

	// state doesn't exist in cache, interact with world ...
	String sState = SendReceiveDirect (sAction, _sAction);
	if ("" == sState)
		return NULL;

	CacheNode* pNewNode = p_Cache->CreateNode (sState, false);
	// pNewNode->s_NormalizedState = Cache::ConvertToNormalizedState (sState);
	_pdqNewNodes->push_back (pNewNode);

	return pNewNode;
}




