#include "ili_cache.h"

// j#define FEATURES_TO_SKIP	3
#define FEATURES_TO_SKIP	0



//															
FeatureBitVector::FeatureBitVector (void)
{
	Clear ();
}


//															
void FeatureBitVector::Clear (void)
{
	c_Features [0] = 0;
	c_Features [1] = 0;
	c_Features [2] = 0;
}


//															
void FeatureBitVector::SetFeatures (char* _zData)
{
	Clear ();

	zchar_dq_t dqFeatures = String::DestructiveSplit (_zData, '\x01');
	assert (FEATURE_COUNT + FEATURES_TO_SKIP == dqFeatures.size ());

	int iBitIndex = 1;
	int iByteIndex = 0;
	for (int i = 0; i < FEATURE_COUNT; ++ i)
	{
		int iValue = atoi (dqFeatures [i + FEATURES_TO_SKIP]);
		c_Features [iByteIndex] |= (unsigned char) ((iValue > 0)? iBitIndex : 0);

		iBitIndex <<= 1;
		if (0 != (0x100 & iBitIndex))
		{
			iBitIndex = 1;
			++ iByteIndex;
		}
		assert (0 != (0xFF | iBitIndex));
	}

	// --- TEST CODE START ------------------------------
	#ifndef NDEBUG
	for (int i = 0; i < FEATURE_COUNT; ++ i)
	{
		int iValue = atoi (dqFeatures [i + FEATURES_TO_SKIP]);
		iValue = (iValue > 0)? 1 : 0;
		if (iValue != operator[] (i))
			cout << iValue << " != " << operator[] (i) << endl;
		assert (iValue == operator[] (i));
	}
	#endif
	// --- TEST CODE END --------------------------------
}



//															
void FeatureBitVector::SetFeature (int _iIndex, int _iValue)
{
	int iByteIndex = _iIndex / 8;
	int iBitIndex = 1 << (_iIndex % 8);

	if (0 == _iValue)
		c_Features [iByteIndex] &= ~iBitIndex;
	else if (1 == _iValue)
		c_Features [iByteIndex] |= iBitIndex;
	else
		assert (false);
}



//															
void FeatureBitVector::GetFeatures (Feature_vec_t& _rvecFeatures)
{
	assert (FEATURE_COUNT == _rvecFeatures.Size ());

	int iBitIndex = 1;
	int iByteIndex = 0;
	for (int i = 0; i < FEATURE_COUNT; ++ i)
	{
		_rvecFeatures [i] = (0 == (c_Features [iByteIndex] & iBitIndex))? (char)0 : (char)1;

		iBitIndex <<= 1;
		if (0 != (0x100 & iBitIndex))
		{
			iBitIndex = 1;
			++ iByteIndex;
		}
		assert (0 != (0xFF | iBitIndex));
	}
}



//															
const int FeatureBitVector::operator[] (int _iIndex) const
{
	int iByteIndex = _iIndex / 8;
	int iBitIndex = 1 << (_iIndex % 8);

	return (0 == (c_Features [iByteIndex] & iBitIndex))? 0 : 1;
}



//															
bool FeatureBitVector::operator== (const FeatureBitVector& _rOther)
{
	return ((c_Features [0] == _rOther.c_Features [0]) &&
			(c_Features [1] == _rOther.c_Features [1]) &&
			(c_Features [2] == _rOther.c_Features [2]));
}


//															
ostream& operator<< (ostream& _rStream, const FeatureBitVector& _rFeatures)
{
	for (int i = 0; i < FEATURE_COUNT; ++ i)
		_rStream << _rFeatures [i] << ' ';
	return _rStream;
}


//															
String& operator<< (String& _rString, const FeatureBitVector& _rFeatures)
{
	for (int i = 0; i < FEATURE_COUNT; ++ i)
		_rString << _rFeatures [i] << ' ';
	return _rString;
}
