#include "EnvInterface.h"
#include <nlp_config.h>
#include <nlp_macros.h>
#include <nlp_time.h>
#include <assert.h>
#include <string.h>
#include "Common.h"



bool	ThreadedEnvironmentInterface::b_ShowComms = false;
bool	ThreadedEnvironmentInterface::b_ShowCommsVerbose = false;


//														
ThreadedEnvironmentInterface::ThreadedEnvironmentInterface (EnvironmentConnectionManager* _pManager)
{
	p_Manager = _pManager;
	p_Connection = NULL;
}



//														
ThreadedEnvironmentInterface::~ThreadedEnvironmentInterface (void)
{
	if (NULL != p_Connection)
		p_Manager->ReleaseConnection (this, p_Connection);
	p_Connection = NULL;

	assert ((EnvironmentConnectionManager*) 0x1 != p_Manager);
	p_Manager = (EnvironmentConnectionManager*) 0x1;
}


//														
bool ThreadedEnvironmentInterface::SendReceive (String& _rMessage, String* _pResponse)
{
	if (NULL == p_Connection)
	{
		*_pResponse = "[ERROR] comm failed\x10\n";
		return false;
	}

	if (true == b_ShowComms)
	{
		String sMsgStripped (_rMessage);
		sMsgStripped.Strip ();
		cout << "   [W <  ] ->" << sMsgStripped << "<-" << endl;
	}

	p_Connection->soc_Agent.SetStandalone (true);
	p_Connection->soc_Agent.SendBlocking ((const char*)_rMessage, _rMessage.length ());
	if (false == p_Connection->soc_Agent.IsConnected ())
	{
		p_Connection->OnError ("Connection failed on send");
		p_Connection = NULL;
		if (true == b_ShowComms)
			cout << "   [W <  ] connection failed on send" << endl;

		*_pResponse = "[ERROR] comm failed\x10\n";
		return false;
	}

	String sResponse;

	int iIdleLoops = 0;
	bool bIdleLoopsHit = false;
	char zData [4097];
	zData [0] = '\0';
	while (NULL == strchr (zData, '\x10'))
	{
		long lBytes = p_Connection->soc_Agent.ReceiveBlocking (zData, 4096, 100);

		// We can loose the connection while waiting for a reply.	
		// In this case, the connection manager will remove this	
		// connection from its pool, and set p_Connection to NULL.	
		if (NULL == p_Connection)
		{
			*_pResponse = "[ERROR] comm failed\x10\n";
			return false;
		}

		if (0 == lBytes)
		{
			p_Connection->OnError ("Connection failed on receive");
			p_Connection = NULL;
			if (true == b_ShowComms)
				cout << "   [W <  ] connection failed on send" << endl;

			*_pResponse = "[ERROR] comm failed\x10\n";
			return false;
		}

		if (lBytes > 0)
		{
			zData [lBytes] = '\0';
			sResponse << zData;
			if (true == b_ShowCommsVerbose)
			{
				if (iIdleLoops > 0)
					cout << "\n";
				cout << "   [W >  ] ->" << zData << "<-" << endl;
			}
			iIdleLoops = 0;
		}
		else
		{
			bIdleLoopsHit = true;
			++ iIdleLoops;
			if (true == b_ShowComms)
				cout << '.' << flush;
		}

		if (iIdleLoops == 500)
		{
			p_Connection->soc_Agent.SendBlocking ("??\n", 3);
			if (true == b_ShowComms)
				cout << '|' << flush;
		}
		if (iIdleLoops > 1000)
		{
			if (true == b_ShowComms)
				cout << "\n   [  C  ] no response from world after 100 seconds" << endl;

			*_pResponse = "[ERROR] comm failed\x10\n";
			return false;
		}
	}
	// p_Connection->soc_Agent.SetStandalone (false);

	sResponse.Strip ();
	*_pResponse = sResponse;

	if ((true == b_ShowComms) && (false == b_ShowCommsVerbose))
	{
		if (true == bIdleLoopsHit)
			cout << "\n";
		String sMsgStripped (_rMessage);
		sMsgStripped.Strip ();
		cout << "   [W >  ] response received for ->" << sMsgStripped << "<-" << endl;
	}

	if (true == sResponse.StartsWith ("[ERROR]"))
		return false;
	return true;
}




//														
bool ThreadedEnvironmentInterface::StartEpisode (String& _rStartingState)
{
	p_Connection = p_Manager->GetConnection (this);

	s_LastEnvironmentId = "";
	int iRetry = 0;
	while (iRetry < 5)
	{
		if ((true == b_ShowComms) || (iRetry > 0))
			cout << "   [W <  ] start episode (" << _rStartingState << ")  " << flush;
		if (true == p_Connection->StartEpisode (_rStartingState, (iRetry > 0)))
		{
			s_LastEnvironmentId << p_Connection->soc_Agent.GetServerId ()
								<< ':' << p_Connection->soc_Agent.GetServiceId ();

			if (p_Connection->s_LastStartState == _rStartingState)
				++ p_Connection->i_SameStartStateCount;
			else
				p_Connection->i_SameStartStateCount = 0;
			p_Connection->s_LastStartState = _rStartingState;
			return true;
		}

		if (NULL == p_Connection)
			return false;
		++ iRetry;
		cerr << "[ERROR]  Attempt to start environment episode failed.  Retrying : " << iRetry << endl;
	}
	cerr << "[ERROR]  Failed to start environment episode after 5 retries." << endl;
	return false;
}



//														
void ThreadedEnvironmentInterface::EndEpisode (void)
{
	if (true == b_ShowComms)
		cout << "   [  C  ] end episode\n" << endl;

	p_Connection->b_Restarted = false;

	if (NULL != p_Connection)
		p_Manager->ReleaseConnection (this, p_Connection);
	p_Connection = NULL;
}



//														
String ThreadedEnvironmentInterface::GetEnvironmentId (void)
{
	return s_LastEnvironmentId;
}



//														
void EnvironmentConnectionSocket::OnDisconnect (void)
{
	if (NULL == p_Parent)
		return;
	String sError;
	sError << "Lost connection to " << s_Name;
	p_Parent->OnError (sError);
	p_Parent = NULL;
}




//														
EnvironmentConnection::EnvironmentConnection (void)
{
	p_Manager = NULL;
	l_Usage = 0;
	b_Restarted = false;
	i_SameStartStateCount = 0;
	pthread_mutex_init (&mtx_Shutdown, NULL);
}


//														
EnvironmentConnection::~EnvironmentConnection (void)
{
	assert ((EnvironmentConnectionManager*)0x1 != p_Manager);
	p_Manager = (EnvironmentConnectionManager*)0x1;
	pthread_mutex_destroy (&mtx_Shutdown);
}


//														
bool EnvironmentConnection::Init (EnvironmentConnectionManager* _pManager,
									String& _rName, 
									String& _rServer,
									String& _rAgentService, 
									String& _rResetService)
{
	p_Manager = _pManager;
	s_Name = _rName;

	soc_Agent.SetStandalone (true);
	if (false == soc_Agent.ConnectBlocking (_rServer, _rAgentService))
		return false;
	soc_Reset.SetStandalone (true);
	if (false == soc_Reset.ConnectBlocking (_rServer, _rResetService))
	{
		soc_Agent.Close ();
		return false;
	}

	soc_Agent.s_Name = "agent";
	soc_Agent.p_Parent = this;
	soc_Reset.s_Name = "reset";
	soc_Reset.p_Parent = this;

	return true;
}


//									//														
bool EnvironmentConnection::StartEpisode (String& _rStartingState, bool _bRetry)
{
	if ((s_LastStartState == _rStartingState) &&
		(true == b_Restarted))
	{
		if (true == ThreadedEnvironmentInterface::b_ShowComms)
			cout << endl;
		return true;
	}

	if ((true == ThreadedEnvironmentInterface::b_ShowComms) || (true == _bRetry))
		cout << soc_Agent.GetServerId () << endl;

	// tell OS agent to drop connection to tcp-relay ...
	soc_Agent.SetStandalone (true);
	soc_Agent.SendBlocking ("%\n", 2);
	sleep (1);

	// send VM reset message to VM agent ...			
	String sVMReset;
	sVMReset << "<< " << _rStartingState << "\n";
	soc_Reset.SendBlocking ((const char*)sVMReset, sVMReset.length ());


	// wait for 'ready' message from OS agent ...		
	Time tmTimer;
	tmTimer.StartTimer ();
	bool bAgentReady = false;
	String sResponse;
	while (true)
	{
		char zData [4096];
		zData [0] = '\0';
		long lBytes = soc_Agent.ReceiveBlocking (zData, 4096, 100);

		// We can loose the connection while waiting for a reply.	
		// In this case, the connection manager will remove this	
		// connection from its pool, and set p_Connection to NULL.	
		// if (NULL == p_Connection)
		//	return false;

		if (0 == lBytes)
		{
			cout << endl;
			cerr << "[WARNING]  Connection to agent failed on receive." << endl;
			OnError ("Connection to agent failed on receive");
			// p_Connection = NULL;
			return false;
		}
		if (lBytes > 0)
		{
			zData [lBytes] = '\0';
			sResponse << zData;
		}

		// check if this is the 'ready' message ...
		if (true == sResponse.Has ("restart complete"))
			bAgentReady = true;

		// we must be sure to eat up the complete 'ready' message ...
		if ((true == bAgentReady) && (true == sResponse.Has ("\x10")))
			break;

		if (true == ThreadedEnvironmentInterface::b_ShowComms)
			cout << '.' << flush;
		if (tmTimer.MinutesToNow () >= 3)
		{
			cout << endl;
			cerr << "[WARNING]  Timing out on environment after 5 minutes." << endl;
			return false;
		}
	}

	if (true == ThreadedEnvironmentInterface::b_ShowComms)
		cout << endl;
	b_Restarted = true;
	return true;
}



					
void EnvironmentConnection::OnError (const char* _zError)
{
	// It is possible for both child sockets (agent & reset)
	// to simultaneously encounter errors and call this		
	// method.  In this case, we would have a race condition
	// since this method deletes the EnvironmentConnection	
	// object.												
	pthread_mutex_lock (&mtx_Shutdown);	// lock !

	EnvironmentConnectionManager* pManager = p_Manager;
	if (NULL == p_Manager)
		return;
	p_Manager = NULL;

	cerr << "[ERROR]  " << _zError << endl;

	soc_Agent.p_Parent = NULL;
	soc_Agent.Close ();
	soc_Reset.p_Parent = NULL;
	soc_Reset.Close ();

	// We need to unlock before calling CloseConnection		
	// below since that's going to delete this object...	
	pthread_mutex_unlock (&mtx_Shutdown);	// unlock !

	// The funcky code below is to avoid loop-back calls.	
	// This code checks if p_Manager is NULL on top to 		
	// OnError calls to a deleted EnvironmentConnection 	
	// object.  But we could get OnError calls as a result	
	// of processing triggered by the CloseConnection call.	
	// So we need to set p_Manager = NULL before that call.	
	pManager->CloseConnection (this);
}







//														
EnvironmentConnectionManager::EnvironmentConnectionManager (void)
{
	i_PendingRequestCount = 0;
	pthread_mutex_init (&mtx_Access, NULL);
}




//														
EnvironmentConnectionManager::~EnvironmentConnectionManager (void)
{
	Shutdown ();
	pthread_mutex_destroy (&mtx_Access);
}



//														
void EnvironmentConnectionManager::CloseConnection (EnvironmentConnection* _pConnection)
{
	pthread_mutex_lock (&mtx_Access);	// lock !

	cerr << "[WARNING]  disconnecting on error from " << _pConnection->s_Name << endl;

	// inform ThreadedEnvironmentInterface that its	
	// connection is dead ...						
	EnvironmentConnectionToThreadedEnvironmentInterface_map_t::iterator	iteInterface;
	iteInterface = map_ConnectionToInterface.find (_pConnection);
	assert (map_ConnectionToInterface.end () != iteInterface);
	ThreadedEnvironmentInterface* pInterface = iteInterface->second;
	pInterface->p_Connection = NULL;

	
	// cleanup connection stores ...
	assert (set_AllConnections.end () != set_AllConnections.find (_pConnection));
	set_AllConnections.erase (_pConnection);
	set_FreeConnections.erase (_pConnection);
	map_ConnectionToInterface.erase (_pConnection);

	delete _pConnection;

	cout << "[INFO]  available worlds : " << set_AllConnections.size () << endl;

	pthread_mutex_unlock (&mtx_Access);	// unlock !
}


//														
void EnvironmentConnectionManager::Init (void)
{
	assert (true == set_ActiveInterfaces.empty ());
	assert (true == map_ConnectionToInterface.empty ());
	assert (true == set_FreeConnections.empty ());

	MaintainConnections (true);
}


//														
void EnvironmentConnectionManager::MaintainConnections (bool _bVerbose)
{
	pthread_mutex_lock (&mtx_Access);	// lock !
	
	// get list of currently active connections ...	
	set <int>	setActiveConnections;
	ITERATE (EnvironmentConnection_set_t, set_AllConnections, ite)
	{
		EnvironmentConnection* pConnection = *ite;
		setActiveConnections.insert (pConnection->i_Index);
	}


	// connect to worlds not already connected to	
	int iConnections = (int)(config)"world_count";
	for (int i = 0; i < iConnections; ++ i)
	{
		if (setActiveConnections.end () != setActiveConnections.find (i))
			continue;

		String sConfigPrefix;
		sConfigPrefix << "world_" << i;
		String sServer ((config)(sConfigPrefix + "_server"));
		String sAgentService ((config)(sConfigPrefix + "_agent_service"));
		String sResetService ((config)(sConfigPrefix + "_reset_service"));

		EnvironmentConnection* pConnection = new EnvironmentConnection;
		pConnection->i_Index = i;
		String sConnectionName;
		sConnectionName << "world " << i << " [" << sServer << ':' 
						<< sAgentService << ':' << sResetService << ']';


		if (true == pConnection->Init (this, 
									   sConnectionName, 
									   sServer,
									   sAgentService,
									   sResetService))
		{
			set_AllConnections.insert (pConnection);
			set_FreeConnections.insert (pConnection);

			cout << "[INFO] Connected to world " << i 
				 << " at [" << sServer << ':' << sAgentService << ':' << sResetService
				 << "]." << endl;
		}
		else if (true == _bVerbose)
			cout << "[WARNING] Failed to connected to world " << i 
				 << " at [" << sServer << ':' << sAgentService << ':' << sResetService
				 << "]." << endl;

	}


	pthread_mutex_unlock (&mtx_Access);	// unlock !
}




//														
void EnvironmentConnectionManager::Shutdown (void)
{
	pthread_mutex_lock (&mtx_Access);	// lock !


	ITERATE (ThreadedEnvironmentInterface_set_t, set_ActiveInterfaces, ite)
	{
		ThreadedEnvironmentInterface* pInterface = *ite;
		pInterface->p_Connection = NULL;
	}
	set_ActiveInterfaces.clear ();
	map_ConnectionToInterface.clear ();

	ITERATE (EnvironmentConnection_set_t, set_AllConnections, ite)
	{
		EnvironmentConnection* pConnection = *ite;
		pConnection->soc_Agent.Close ();
		pConnection->soc_Reset.Close ();
		delete pConnection;
	}
	set_AllConnections.clear ();
	set_FreeConnections.clear ();


	pthread_mutex_unlock (&mtx_Access);	// unlock !
}




//														
EnvironmentConnection* EnvironmentConnectionManager::GetConnection (ThreadedEnvironmentInterface* _pInterface)
{
	// should we add a watchdog timer to this so we don't get stuck forever?
	pthread_mutex_lock (&mtx_Access);		// lock 
	++ i_PendingRequestCount;
	pthread_mutex_unlock (&mtx_Access);		// unlock !
	
	while (true)
	{
		pthread_mutex_lock (&mtx_Access);		// lock 

		if (false == set_FreeConnections.empty ())
		{
			long l_MinUsage = -1;
			EnvironmentConnection* pConnectionToUse = NULL;
			EnvironmentConnection_set_t::iterator	iteToUse;
			ITERATE (EnvironmentConnection_set_t, set_FreeConnections, ite)
			{
				EnvironmentConnection* pConnection = *ite;
				if ((-1 == l_MinUsage) || (l_MinUsage > pConnection->l_Usage))
				{
					l_MinUsage = pConnection->l_Usage;
					iteToUse = ite;
					pConnectionToUse = pConnection;
				}
			}
			++ pConnectionToUse->l_Usage;

			set_FreeConnections.erase (iteToUse);
			set_ActiveInterfaces.insert (_pInterface);
			map_ConnectionToInterface.insert (make_pair (pConnectionToUse, _pInterface));
			-- i_PendingRequestCount;
			
			pthread_mutex_unlock (&mtx_Access);	// unlock 
			return pConnectionToUse;
		}

		pthread_mutex_unlock (&mtx_Access);		// unlock !

		usleep (10000);
	}
}




//														
void EnvironmentConnectionManager::ReleaseConnection (ThreadedEnvironmentInterface* _pInterface,
														EnvironmentConnection* _pConnection)
{
	pthread_mutex_lock (&mtx_Access);	// lock !
	
	assert (set_AllConnections.end () != set_AllConnections.find (_pConnection));
	assert (set_FreeConnections.end () == set_FreeConnections.find (_pConnection));
	set_FreeConnections.insert (_pConnection);

	assert (set_ActiveInterfaces.end () != set_ActiveInterfaces.find (_pInterface));
	set_ActiveInterfaces.erase (_pInterface);

	assert (map_ConnectionToInterface.end () != map_ConnectionToInterface.find (_pConnection));
	map_ConnectionToInterface.erase (_pConnection);

	pthread_mutex_unlock (&mtx_Access);	// unlock !
}



//														
bool EnvironmentConnectionManager::PendingRequestPresent (void)
{
	bool bRequestPresent = false;

	pthread_mutex_lock (&mtx_Access);
	bRequestPresent = (i_PendingRequestCount > 0);
	pthread_mutex_unlock (&mtx_Access);

	return bRequestPresent;
}



//														
void EnvironmentConnectionManager::PreemptiveEnvironmentReset (void)
{
	// find connection that can be preemptively	
	// restarted ...							
	pthread_mutex_lock (&mtx_Access);			// lock 
	if (true == set_FreeConnections.empty ())
	{
		pthread_mutex_unlock (&mtx_Access);		// unlock !
		return;
	}

	EnvironmentConnection* pConnectionToReset = NULL;
	ITERATE (EnvironmentConnection_set_t, set_FreeConnections, ite)
	{
		EnvironmentConnection* pConnection = *ite;
		if ((pConnection->i_SameStartStateCount < 1) ||
			(true == pConnection->b_Restarted))
			continue;

		pConnectionToReset = pConnection;
		set_FreeConnections.erase (ite);
		break;
	}
	pthread_mutex_unlock (&mtx_Access);			// unlock !

	if (NULL == pConnectionToReset)
		return;

	// perform preemtive restart ...			
	pConnectionToReset->StartEpisode (pConnectionToReset->s_LastStartState, false);

	pthread_mutex_lock (&mtx_Access);			// lock 
	set_FreeConnections.insert (pConnectionToReset);
	pthread_mutex_unlock (&mtx_Access);			// unlock !
}



//														
void* EnvironmentConnectionManager::PreemptiveEnvironmentReset (void* _pArg)
{
	EnvironmentConnectionManager* pManager = (EnvironmentConnectionManager*) _pArg;
	while (true)
	{
		if (false == pManager->PendingRequestPresent ())
			pManager->PreemptiveEnvironmentReset ();
		usleep (100000);
	};
	return NULL;
}


