#ifndef __ENV_INTERFACE_
#define __ENV_INTERFACE_

#include <set>
#include <map>
#include <deque>
#include <nlp_string.h>
#include <nlp_socket.h>
#include <ili_cache.h>
#include <pthread.h>


class EnvironmentConnection;
class EnvironmentConnectionManager;
class ThreadedEnvironmentInterface;
typedef set <EnvironmentConnection*>			EnvironmentConnection_set_t;
typedef set <ThreadedEnvironmentInterface*>		ThreadedEnvironmentInterface_set_t;
typedef map <EnvironmentConnection*, ThreadedEnvironmentInterface*>	
												EnvironmentConnectionToThreadedEnvironmentInterface_map_t;



// -----------------------------------------------------
class ThreadedEnvironmentInterface : public EnvironmentInterface
{
	friend class EnvironmentConnectionManager;

	protected:
		EnvironmentConnectionManager*	p_Manager;
		EnvironmentConnection*			p_Connection;
		String							s_LastEnvironmentId;

	public:
		static bool						b_ShowComms;
		static bool						b_ShowCommsVerbose;

		ThreadedEnvironmentInterface (EnvironmentConnectionManager* _pManager);
		~ThreadedEnvironmentInterface (void);

		// void Close (void);

		virtual bool SendReceive (String& _rMessage, String* _pResponse);
		virtual bool StartEpisode (String& _rStartingState);
		virtual void EndEpisode (void);
		virtual String GetEnvironmentId (void);
};




// -----------------------------------------------------
class EnvironmentConnectionSocket : public ClientSocket
{
	friend class EnvironmentConnection;

	private:
		String							s_Name;
		EnvironmentConnection*			p_Parent;

	public:
		virtual void OnDisconnect (void);
};



// -----------------------------------------------------
class EnvironmentConnection
{
	private:
		EnvironmentConnectionManager*	p_Manager;
		pthread_mutex_t					mtx_Shutdown;

	public:
		EnvironmentConnectionSocket		soc_Agent;
		EnvironmentConnectionSocket		soc_Reset;
		String							s_Name;
		int								i_Index;
		long							l_Usage;
		String							s_LastStartState;
		int								i_SameStartStateCount;
		bool							b_Restarted;


		EnvironmentConnection (void);
		~EnvironmentConnection (void);

		bool Init (EnvironmentConnectionManager* _pManager,
					String& _rName, 
					String& _rServer,
					String& _rAgentService, 
					String& _rResetService);

		bool StartEpisode (String& _rStartingState, bool _bRetry);

		void  OnError (const char* _zError);
};




// -----------------------------------------------------
class EnvironmentConnectionManager
{
	friend class EnvironmentConnection;

	private:
		pthread_mutex_t		mtx_Access;

	protected:
		EnvironmentConnection_set_t			set_AllConnections;
		EnvironmentConnection_set_t			set_FreeConnections;
		ThreadedEnvironmentInterface_set_t	set_ActiveInterfaces;
		EnvironmentConnectionToThreadedEnvironmentInterface_map_t
											map_ConnectionToInterface;
		int									i_PendingRequestCount;

		void CloseConnection (EnvironmentConnection* _pConnection);

	public:
		pthread_t			thr_PreemptiveReset;

		EnvironmentConnectionManager (void);
		~EnvironmentConnectionManager (void);

		void Init (void);
		void MaintainConnections (bool _bVerbose = false);
		void Shutdown (void);

		EnvironmentConnection* GetConnection (ThreadedEnvironmentInterface* _pInterface);
		void ReleaseConnection (ThreadedEnvironmentInterface* _pInterface,
								EnvironmentConnection* _pConnection);

		bool PendingRequestPresent (void);
		void PreemptiveEnvironmentReset (void);
		static void* PreemptiveEnvironmentReset (void* _pArg);
};



#endif
