#include "Server.h"
#include <assert.h>
#include <string.h>
#include <nlp_time.h>
#include "Common.h"

Cache* 							Server::sp_Cache = NULL;
EnvironmentConnectionManager*	Server::sp_EnvConManager = NULL;
bool							Server::b_ShowComms = false;
bool							Server::b_ShowCommsVerbose = false;


//												
Server::Server (void)
{
	p_EnvironmentInterface = NULL;
	b_ServerActive = false;
	i_ResponseDelay = 0;
}


Server::~Server (void)
{
	assert ((ThreadedEnvironmentInterface*)0x1 != p_EnvironmentInterface);
	if (NULL != p_EnvironmentInterface)
		delete p_EnvironmentInterface;
	p_EnvironmentInterface = (ThreadedEnvironmentInterface*)0x1;
}



//												
void Server::Init (Cache* _pCache, EnvironmentConnectionManager* _pManager)
{
	sp_Cache = _pCache;
	sp_EnvConManager = _pManager;
}



//												
ClientSocket* Server::CreateClient (void)
{
	// create server object ...				
	Server* pServer = new Server;
	pServer->SetStandalone (true);
	pServer->p_EnvironmentInterface = new ThreadedEnvironmentInterface (sp_EnvConManager);
	pServer->o_CacheInterface.Init (sp_Cache, pServer->p_EnvironmentInterface);
	pServer->b_ServerActive = true;

	// create thread for server object ...	
	pthread_create (&pServer->thr_Server, NULL, RunThread, pServer);
	pthread_detach (pServer->thr_Server);

	return pServer;
}



//												
void Server::OnConnect (void)
{
	cout << "[INFO]  Connected to [" << s_Server << ':' << s_Service << ']' << endl;
	SetNoDelay (true);
}



//												
void Server::OnDisconnect (void)
{
	o_CacheInterface.EndEpisode ();
	// p_EnvironmentInterface->EndEpisode ();	// this line must have been wrong ...
	cout << "[INFO]  Disconnected from [" << s_Server << ':' << s_Service << ']' << endl;
	b_ServerActive = false;
}



//												
void Server::OnReceive (const void* _zData, long _lBytes)
{
	o_Data.Append (_zData, _lBytes);

	while (true == o_Data.HasTerminator ('\n'))
	{
		String sMessage = o_Data.PopFirstMessageAsString ('\n');
		char cType = sMessage.at (0);

		// Delay responses if a delay time has been	
		// set. This is to support slow clients who	
		// might miss the response otherwise...		
		if (i_ResponseDelay > 0)
			// delay is set in milliseconds	
			usleep (1000 * i_ResponseDelay);


		// manually triggered support features ...	
		if ("help\n" == sMessage)
		{
			String sResponse;
			sResponse << "Valid support requests:\n"
						 "  help\n"
						 "  save cache\n"
						 "  show/hide learner comms\n"
						 "  learner comms verbose\n"
						 "  show/hide world comms\n"
						 "  world comms verbose\n";
			SendBlocking ((const char*)sResponse, sResponse.length ());
		}

		else if ("save cache\n" == sMessage)
		{
			String sResponse;
			if (NULL == sp_Cache)
				sResponse = "[ERROR] Failed to save cache. Pointer to cache is NULL\n";
			else
			{
				sp_Cache->Save ();
				sResponse = "[INFO] Cache saved\n.";
			}
			SendBlocking ((const char*)sResponse, sResponse.length ());
		}

		else if ("show learner comms\n" == sMessage)
			Server::b_ShowComms = true;
		else if ("hide learner comms\n" == sMessage)
		{
			Server::b_ShowComms = false;
			Server::b_ShowCommsVerbose = false;
		}
		else if ("learner comms verbose\n" == sMessage)
			Server::b_ShowCommsVerbose = true;
		else if ("show world comms\n" == sMessage)
			ThreadedEnvironmentInterface::b_ShowComms = true;
		else if ("hide world comms\n" == sMessage)
		{
			ThreadedEnvironmentInterface::b_ShowComms = false;
			ThreadedEnvironmentInterface::b_ShowCommsVerbose = false;
		}
		else if ("world comms verbose\n" == sMessage)
			ThreadedEnvironmentInterface::b_ShowCommsVerbose = true;


		else if ("[known world only]\n" == sMessage)
		{
			String sResponse ("Known-world-only mode active\x10\n");
			SendBlocking ((const char*)sResponse, sResponse.length ());
			cout << "[INFO]  Switching to known-world-only mode." << endl;
			o_CacheInterface.b_KnownWorldOnly = true;
		}

		else if (true == sMessage.StartsWith ("[delay response]"))
		{
			String sDelay = sMessage.substr (strlen ("[delay response]"));
			i_ResponseDelay = sDelay;
			if (i_ResponseDelay < 0)
			{
				cout << "[ERROR]  Response delay must be positive. Requested : " 
					 << i_ResponseDelay << endl;
				i_ResponseDelay = 0;
			}
			else if (0 == i_ResponseDelay)
				cout << "[INFO]  Response delay removed." << endl;
			else
				cout << "[INFO]  Delaying responses by " << i_ResponseDelay 
					 << " milliseconds." << endl;
		}

		// start episode ...						
		else if ('>' == cType)
		{
			String sRequiredStartState (sMessage.substr (1));
			sRequiredStartState.Strip ();
			if (true == b_ShowComms)
				cout << "   [  < L] start episode (" << sRequiredStartState << ")" << endl;
			if (true == o_CacheInterface.StartEpisode (sRequiredStartState))
			{
				String sResponse ("restart complete\x10\n");
				SendBlocking ((const char*)sResponse, sResponse.length ());
				if (true == b_ShowComms)
					cout << "   [  > L] restart complete" << endl;
			}
			else
			{
				String sResponse ("[ERROR] restart failed\x10\n");
				SendBlocking ((const char*)sResponse, sResponse.length ());
				if (true == b_ShowComms)
					cout << "   [  > L] restart failed!" << endl;
			}
		}

		// end episode ...							
		else if ('.' == cType)
		{
			if (true == b_ShowComms)
				cout << "   [  < L] end episode" << endl;
			o_CacheInterface.EndEpisode ();
		}

		// environment interaction ...				
		else if (('*' == cType) || ('!' == cType))
		{
			if (true == b_ShowComms)
			{
				String sMsgStripped (sMessage);
				sMsgStripped.Strip ();
				cout << "   [  < L] ->" << sMsgStripped << "<-" << endl;
			}
			String sResponse = o_CacheInterface.SendReceive (sMessage);

			// check for errors...			
			if ("" == sResponse)
			{
				String sMsgStripped (sMessage);
				sMsgStripped.Strip ();
				cout << "[WARNING] Caught empty response from CacheInterface::SendReceive for message ["
					 << sMsgStripped << ']' << endl;
				sResponse = "[ERROR] unknown error\x10\n";
			}
			if (('!' == cType) && ("restart complete.\x10\n" == sResponse))
			{
				cout << "[INFO] Caught 'restart complete' response for '!...' . "
						"       Replacing with 'action completed, env has changed." << endl;
				sResponse = "action completed, env has changed.\x10\n";
			}

			// send response ...			
			SendBlocking ((const char*)sResponse, sResponse.length ());
			if (true == b_ShowComms)
			{
				if (true == b_ShowCommsVerbose)
					cout << "   [  > L] ->" << sResponse << "<-" << endl;
				else
				{
					String sMsgStripped (sMessage);
					sMsgStripped.Strip ();
					cout << "   [  > L] response sent for ->" << sMsgStripped << "<-\n";
				}
			}
		}

		
		// unknown request ...						
		else
		{
			cerr << "[WARNING] Unknown message received ->" << sMessage << "<-" << endl;
			String sResponse ("[[ERROR]] Unknown request\x10\n");
			SendBlocking ((const char*)sResponse, sResponse.length ());
		}
	}
}



//												
void* Server::RunThread (void* _pArg)
{
	Server* pServer = (Server*) _pArg;
	while (true == pServer->b_ServerActive)
	{
		pServer->ProcessEvents (1000);
	};
	return NULL;
}




