#include <iostream>
#include <time.h>
#include <signal.h>
#include <nlp_config.h>
#include <ili_cache.h>
#include "Server.h"
#include "EnvInterface.h"
// #include "Cache.h"
#include "Common.h"


void OnSignal (int _iSignal)
{
	cout << "closing all sockets on termination." << endl;
	AllSockets::CloseAndDeleteAllSockets ();
	exit (0);
}


int main (int argc, const char* argv[])
{
	signal (SIGINT, OnSignal);

	Config::SetCommandLine (argc, argv);

	// initialize cache ...	
	cout << "initializing cache ..." << endl;
	EnvObject::i_ObjectMemorySize = sizeof (EnvObject);
	Cache oCache;
	oCache.Load ();
	cout << "   will save cache every " << (config)"cache_save_period" << " minutes." << endl;

	CacheInterface::i_ForegroundFeatureIndex = (config)"foreground_feature_index";
	CacheInterface::i_NewlyForegroundFeatureIndex = (config)"newly_foreground_feature_index";
	CacheInterface::b_Verbose = (1 == (int)(config)"verbose");


	// initialize interface to environment ... 
	cout << "initializing environment connection manager ..." << endl;
	EnvironmentConnectionManager oManager;
	oManager.Init ();
	pthread_create (&oManager.thr_PreemptiveReset, 
					NULL, 
					EnvironmentConnectionManager::PreemptiveEnvironmentReset,
					&oManager);
	pthread_detach (oManager.thr_PreemptiveReset);


	// initialize server ... 
	cout << "initializing server ..." << endl;
	Server::Init (&oCache, &oManager);


	// start server socket ... 
	cout << "starting service on [" << (config)"service_port" << "]." << endl;
	ServerSocket socServer;
	socServer.ListenNonBlocking ((config)"service_port",
								 (int)(config)"service_backlog",
								 Server::CreateClient);


	long lCacheSavePeriod = 60 * (long)(config)"cache_save_period";
	time_t tLastSave = time (NULL);

	long lReconnectPeriod = 60 * (long)(config)"world_reconnect_period";
	time_t tLastConnectionCheck = time (NULL);

	while (true)
	{
		AllSockets::ProcessEvents (1000);

		if (lReconnectPeriod <= (time (NULL) - tLastConnectionCheck))
		{
			oManager.MaintainConnections ();
			tLastConnectionCheck = time (NULL);
		}

		if (lCacheSavePeriod <= (time (NULL) - tLastSave))
		{
			oCache.Save ();
			tLastSave = time (NULL);
		}

		// if (Socket::NotConnected == socServer.e_ConnectionStatus)
		// {
		//	socServer.ListenNonBlocking ((config)"service_port",
		//								 (int)(config)"service_backlog",
		//								 Server::CreateClient);
		// }
	}
}
